#ifndef EXPRESSION_H
#define EXPRESSION_H

#include <iostream>
#include <string>
#include <sstream>	//um double in string umzuwandeln.

////////////////////////////////////////////////////////////////////////////////
// Abstrakte Interface Klasse
class Expression
{
public:
	virtual ~Expression(){}
	virtual double evaluate() = 0;
	virtual std::string print() const = 0;
};

////////////////////////////////////////////////////////////////////////////////
// arbeitet als Interface für den Taschenrechner.
class Result : public Expression
{
public:
	Result(Expression* incoming) // Konstruktor
	{
		ausdruck = incoming;
	}

	virtual ~Result() // Destruktor
	{
		delete ausdruck;
		ausdruck = NULL; /* Verhindert den Zugriff auf einen wilden Zeiger.
		Kann bei den anderen Erben entfallen, da die abgerissenen Zeiger bei
		delete ausdruck unerreichbar werden.*/
	}

	virtual std::string print() const // stößt die rekursive Ausgabe an
	{
		return ausdruck->print();
	}
	virtual double evaluate() // stößt die rekursive Berechnung an
	{
		return ausdruck->evaluate();
	}

private:
	Expression* ausdruck;
};

////////////////////////////////////////////////////////////////////////////////
// repräsentiert eine Konstante.
class Const : public Expression
{
public:
	Const(double incoming)
	{
		wert = incoming;
	}

	virtual ~Const(){}; // Const fordert keinen Speicher auf dem Heap an

	virtual double evaluate()
	{
		return wert;
	}

	virtual std::string print() const
	{
		std::stringstream ss;
		ss << wert;
		return ss.str();
	}
private:
	double wert;
};

////////////////////////////////////////////////////////////////////////////////
// Addiert die Ausrücke operandA und operandB.
class Add : public Expression
/* Die Kommentare in dieser Klasse gelten sinngemäß auch für die weiter
 * unten definierten Klassen.
 */
{
public:
	Add(Expression* inA, Expression* inB)
	/* Standardkonstruktor, legt die Operanden fest. Operanden können
	 * Konstanten oder andere Ausdrücke sein, auf die Zeiger übergeben
	 * wurden.
	 */
	{
		operandA = inA;
		operandB = inB;
	}

	virtual ~Add()
	/* Standarddestruktor, gibt den vom Konstruktor für die Operanden
	 * auf dem Heap angeforderten Speicher wieder frei.
	 * Der Destruktor wird rekursiv aufgerufen, wenn ~Result aufgerufen
	 * wird.
	 */
	{
		delete operandA;
		delete operandB;
	}

	virtual double evaluate()
	/* Stößt die Auswertung der Operation an. Die Auswertung geht rekursiv vor,
	 * bis die Operanden Konstanten sind. Dann wird rekursiv zurückgesprungen.
	 * Die eigentliche Auswertung erfolgt nach dem Rekursionsausstieg.
	 */
	{
		return operandA->evaluate() + operandB->evaluate();
	}

	virtual std::string print() const
	/* Stößt die rekursive Ausgabe des Terms auf der Konsole an. Dazu wird
	 * ähnlich vorgegangen wie bei der Auswertung des Terms.
	 */
	{
		std::stringstream ss;
		ss << "(" << operandA->print() << " + " << operandB->print() << ")";
		return ss.str();;
	}
private:
	/* Die privaten Variablen halten Zeiger auf die beiden zugeordneten
	 * Operanden.
	 */
	Expression* operandA;
	Expression* operandB;
};

////////////////////////////////////////////////////////////////////////////////
// Subtrahiert den Ausdruck operandB von operandA.
class Sub : public Expression
{
public:
	Sub(Expression* inA, Expression* inB)
	{
		operandA = inA;
		operandB = inB;
	}

	virtual ~Sub()
		{
			delete operandA;
			delete operandB;
		}

	virtual double evaluate()
	{
		return operandA->evaluate() - operandB->evaluate();
	}

	virtual std::string print() const
	{
		std::stringstream ss;
		ss << "(" << operandA->print() << " - " << operandB->print() << ")";
		return ss.str();
	}
private:
	Expression* operandA;
	Expression* operandB;
};

////////////////////////////////////////////////////////////////////////////////
// Multipliziert die Ausrücke operandA und operandB.
class Mul : public Expression
{
public:
	Mul(Expression* inA, Expression* inB)
	{
		operandA = inA;
		operandB = inB;
	}
	virtual ~Mul()
		{
			delete operandA;
			delete operandB;
		}

	virtual double evaluate()
	{
		return operandA->evaluate() * operandB->evaluate();
	}

	virtual std::string print() const
	{
		std::stringstream ss;
		ss << "(" << operandA->print() << " * " << operandB->print() << ")";
		return ss.str();
	}
private:
	Expression* operandA;
	Expression* operandB;
};

////////////////////////////////////////////////////////////////////////////////
// Dividiert den Ausrucke operandA durch operandB.
class Div : public Expression
{
public:
	Div(Expression* inA, Expression* inB)
	{
		operandA = inA;
		operandB = inB;
	}
	virtual ~Div()
		{
			delete operandA;
			delete operandB;
		}

	virtual double evaluate()
	{
		return operandA->evaluate() / operandB->evaluate();
	}

	virtual std::string print() const
	{
		std::stringstream ss;
		ss << "(" << operandA->print() << " / " << operandB->print() << ")";
		return ss.str();
	}
private:
	Expression* operandA;
	Expression* operandB;
};
#endif
